@@ -1,17 +1,24 @@
-# TODO: Implement FileSystemS3
+from functools import lru_cache
 import logging
 import os
 import shutil
+from time import time
 import xarray as xr
 from datetime import datetime
-from pydantic import BaseSettings, validator
+from pydantic import BaseSettings, validator, Field
 from pathlib import Path
-from typing import Any, Dict, List, Optional
+from typing import Any, Dict, List, Optional, Union
 from .base import Storage
-from .handlers import FileHandler, ZarrHandler
+from .handlers import FileHandler, NetCDFHandler, ZarrHandler
 from ..utils import get_filename
 
-__all__ = ["FileSystem"]
+import boto3
+import botocore.exceptions
+
+import tempfile
+
+
+__all__ = ["FileSystem", "FileSystemS3", "ZarrLocalStorage"]
 
 # IDEA: interval / split files apart by some timeframe (e.g., 1 day)
 #
@@ -63,15 +70,14 @@ class FileSystem(Storage):
         merge_fetched_data_kwargs: Dict[str, Any] = dict()
 
         @validator("storage_root")
-        @classmethod
         def _ensure_storage_root_exists(cls, storage_root: Path) -> Path:
             if not storage_root.is_dir():
                 logger.info("Creating storage root at: %s", storage_root.as_posix())
                 storage_root.mkdir(parents=True)
             return storage_root
 
-    parameters: Parameters = Parameters()
-    handler: FileHandler
+    parameters: Parameters = Field(default_factory=Parameters)  # type: ignore
+    handler: FileHandler = Field(default_factory=NetCDFHandler)
 
     def save_data(self, dataset: xr.Dataset):
         """-----------------------------------------------------------------------------
@@ -169,6 +175,156 @@ class FileSystem(Storage):
         return anc_datastream_dir / filepath.name
 
 
+class FileSystemS3(FileSystem):
+    """------------------------------------------------------------------------------------
+    Handles data storage and retrieval for file-based data formats in an AWS S3 bucket.
+
+    Args:
+        parameters (Parameters): File-system and AWS-specific parameters, such as the root
+            path to where files should be saved, or additional keyword arguments to
+            specific functions used by the storage API. See the S3Storage.Parameters
+            class for more details.
+        handler (FileHandler): The FileHandler class that should be used to handle data
+            I/O within the storage API.
+
+    ------------------------------------------------------------------------------------"""
+
+    class Parameters(BaseSettings):  # type: ignore
+        storage_root: Path = Field(Path("storage/root"), env="TSDAT_STORAGE_ROOT")
+        """The path on disk where data and ancillary files will be saved to. Defaults to
+        the `storage/root` folder in the top level of the storage bucket."""
+
+        bucket: str = Field("tsdat-storage", env="TSDAT_S3_BUCKET_NAME")
+        """The name of the S3 bucket that the storage class should attach to."""
+
+        region: str = Field("us-west-2", env="AWS_DEFAULT_REGION")
+        """The AWS region of the storage bucket. Defaults to "us-west-2"."""
+
+        merge_fetched_data_kwargs: Dict[str, Any] = dict()
+        """Keyword arguments to xr.merge. Note: this will only be called if the
+        DataReader returns a dictionary of xr.Datasets for a single saved file."""
+
+    parameters: Parameters = Field(default_factory=Parameters)  # type: ignore
+
+    @validator("parameters")
+    def check_authentication(cls, parameters: Parameters):
+        session = FileSystemS3._get_session(
+            region=parameters.region, timehash=FileSystemS3._get_timehash()
+        )
+        try:
+            session.client("sts").get_caller_identity().get("Account")  # type: ignore
+        except botocore.exceptions.ClientError:
+            raise ValueError(
+                "Could not connect to the AWS client. This is likely due to"
+                " misconfigured or expired credentials."
+            )
+        return parameters
+
+    @validator("parameters")
+    def ensure_bucket_exists(cls, parameters: Parameters):
+        session = FileSystemS3._get_session(
+            region=parameters.region, timehash=FileSystemS3._get_timehash()
+        )
+        s3 = session.resource("s3", region_name=parameters.region)  # type: ignore
+        try:
+            s3.meta.client.head_bucket(Bucket=parameters.bucket)
+        except botocore.exceptions.ClientError:
+            logger.warning("Creating bucket '%s'.", parameters.bucket)
+            s3.create_bucket(Bucket=parameters.bucket)
+        return parameters
+
+    @property
+    def session(self):
+        return FileSystemS3._get_session(
+            region=self.parameters.region, timehash=FileSystemS3._get_timehash()
+        )
+
+    @property
+    def bucket(self):
+        s3 = self.session.resource("s3", region_name=self.parameters.region)  # type: ignore
+        return s3.Bucket(name=self.parameters.bucket)
+
+    @staticmethod
+    @lru_cache()
+    def _get_session(region: str, timehash: int = 0):
+        """------------------------------------------------------------------------------------
+        Creates a boto3 Session or returns an active one.
+
+        Borrowed approximately from https://stackoverflow.com/a/55900800/15641512.
+
+        Args:
+            region (str): The session region.
+            timehash (int, optional): A time hash used to cache repeated calls to this
+                function. This should be generated using tsdat.io.storage.get_timehash().
+
+        Returns:
+            boto3.session.Session: An active boto3 Session object.
+
+        ------------------------------------------------------------------------------------"""
+        del timehash
+        return boto3.session.Session(region_name=region)
+
+    @staticmethod
+    def _get_timehash(seconds: int = 3600) -> int:
+        return round(time() / seconds)
+
+    def save_data(self, dataset: xr.Dataset):
+        datastream: str = dataset.attrs["datastream"]
+        filename = get_filename(dataset, self.handler.extension)
+
+        with tempfile.TemporaryDirectory() as tmp_dir:
+            tmp_filepath = Path(tmp_dir) / filename
+            self.handler.writer.write(dataset, tmp_filepath)
+            s3_filepath = self._get_dataset_filepath(dataset, datastream)
+            self.bucket.upload_file(Filename=str(tmp_filepath), Key=str(s3_filepath))
+
+        logger.info(
+            "Saved %s dataset to %s in AWS S3 bucket %s",
+            datastream,
+            str(s3_filepath),
+            self.parameters.bucket,
+        )
+
+    def save_ancillary_file(self, filepath: Path, datastream: str):
+        s3_filepath = self._get_ancillary_filepath(filepath, datastream)
+        self.bucket.upload_file(Filename=str(filepath), Key=str(s3_filepath))
+        logger.info("Saved %s ancillary file to: %s", str(s3_filepath))
+
+    def _find_data(self, start: datetime, end: datetime, datastream: str) -> List[Path]:
+        prefix = str(self.parameters.storage_root / "data" / datastream) + "/"
+        objects = self.bucket.objects.filter(Prefix=prefix)
+        filepaths = [
+            Path(obj.key) for obj in objects if obj.key.endswith(self.handler.extension)
+        ]
+        return self._filter_between_dates(filepaths, start, end)
+
+    def _open_data_files(self, *filepaths: Path) -> List[xr.Dataset]:
+        dataset_list: List[xr.Dataset] = []
+        with tempfile.TemporaryDirectory() as tmp_dir:
+            for s3_filepath in filepaths:
+                tmp_filepath = str(Path(tmp_dir) / s3_filepath.name)
+                self.bucket.download_file(
+                    Key=str(s3_filepath),
+                    Filename=tmp_filepath,
+                )
+                data = self.handler.reader.read(tmp_filepath)
+                if isinstance(data, dict):
+                    data = xr.merge(data.values())  # type: ignore
+                data = data.compute()  # type: ignore
+                dataset_list.append(data)
+        return dataset_list
+
+    def exists(self, key: Union[Path, str]) -> bool:
+        return self.get_obj(str(key)) is not None
+
+    def get_obj(self, key: Union[Path, str]):
+        objects = self.bucket.objects.filter(Prefix=str(key))
+        try:
+            return next(obj for obj in objects if obj.key == str(key))
+        except StopIteration:
+            return None
+
+
 class ZarrLocalStorage(Storage):
     """---------------------------------------------------------------------------------
     Handles data storage and retrieval for zarr archives on a local filesystem.
@@ -195,8 +351,8 @@ class ZarrLocalStorage(Storage):
         the `storage/root` folder in the active working directory. The directory is
         created as this parameter is set, if the directory does not already exist."""
 
-    handler: ZarrHandler = ZarrHandler()
-    parameters: Parameters = Parameters()
+    parameters: Parameters = Field(default_factory=Parameters)
+    handler: ZarrHandler = Field(default_factory=ZarrHandler)
 
     def save_data(self, dataset: xr.Dataset):
         """-----------------------------------------------------------------------------
@@ -258,19 +414,8 @@ class ZarrLocalStorage(Storage):
         return anc_datastream_dir / filepath.name
 
 
-# class S3Storage(FileSystem):
-
-#     class Parameters(FileSystem.Parameters):
-#         bucket: str
-#         region: str = "us-west-2"
-
-#     parameters: Parameters
-
-#     def save_data(self, dataset: xr.Dataset):
-#         return super().save_data(dataset)
-
-#     def fetch_data(self, start: datetime, end: datetime, datastream: str) -> xr.Dataset:
-#         return super().fetch_data(start, end, datastream)
-
-#     def save_ancillary_file(self, filepath: Path, datastream: str):
-#         return super().save_ancillary_file(filepath, datastream)
+# HACK: Update forward refs to get around error I couldn't replicate with simpler code
+# "pydantic.errors.ConfigError: field "parameters" not yet prepared so type is still a ForwardRef..."
+FileSystem.update_forward_refs(Parameters=FileSystem.Parameters)
+FileSystemS3.update_forward_refs(Parameters=FileSystemS3.Parameters)
+ZarrLocalStorage.update_forward_refs(Parameters=ZarrLocalStorage.Parameters)
